#!/usr/bin/env pluginwrapper3

import hashlib
import json
import os
import ssl

from urllib.request import urlopen, Request
from urllib.error import HTTPError
from urllib.parse import urlencode

import time
import binascii
from configparser import ConfigParser
from io import StringIO

from dns import resolver, reversename
from Crypto.Cipher import AES


DEBUG = True

DEFAULT_CONFIG = {
    'provider': 'local-config',
    'domain': 'local',
    'under_group': '',
    'reset_credential': '0',
    'use_credential': 'gateway',
    'username': '',
    'password': '',
    'ca': '',
    'check_host_name': '1'
}
AGENT_SECRET = b'q6pdqy84mzs6ryf1'
AGENT_SCB = b'AQF90ZMqZAwc7_qR'
AGENT_IDENTIFIER = b'wmbns08zb9xucbmp'


class PassmanException(Exception):
    pass


class Passman:
    def __init__(self, config):
        self._access_token = None
        self._agent_auth = 'null'

        self._passman_url = config.get('passman', 'url')
        self._passman_ca_cert = config.get('passman', 'ca')
        self._passman_ca_cert = None if self._passman_ca_cert == '' else self._passman_ca_cert
        self._passman_check_host_name = config.getboolean('passman', 'check_host_name')

    def set_access_token(self, access_token):
        self._access_token = access_token

    def set_auth_token(self, auth_token):
        self._agent_auth = auth_token

    def get_account(self, target_username, target_host, under_group=''):
        target_selector = {'__eqci_name': target_host, '_include_subfolders': True}
        if under_group:
            target_selector['_under_group'] = under_group

        target_ids = self._get('targets', target_selector)
        if len(target_ids) == 0:
            raise PassmanException("No such target_host '{}' under group '{}'".format(target_host, under_group))

        if DEBUG:
            print('target ids', target_ids)

        # maybe get whole account to check if credential type is available
        accounts = self._get('accounts',
                             {'target': target_ids[0], 'name': target_username,
                              '_expand': True, '_expand_target': True, '_include_subfolders': True})
        if len(accounts['accounts']) != 1:
            raise PassmanException("No such target_username '{}'".format(target_username))

        return accounts['accounts'][0]

    def checkout_credential(self, account, credential_type):
        job_data = {
            "account": account['id'],
            "credential_type": credential_type
        }

        is_exclusive = False
        if account['target']['permitted_checkout_types'] == 'exclusive only':
            job_data['exclusive_duration'] = account['target']['max_exclusive_duration']
            is_exclusive = True

        checkout_job = self._post('jobs/LoginCheckoutCredentials', job_data)
        if checkout_job is not None and 'result' in checkout_job and credential_type in checkout_job['result']:
            return checkout_job['result'][credential_type], is_exclusive

        raise PassmanException('Error checking out credential {}'.format(checkout_job))

    def checkin_credentials(self, account_id):
        job_data = {"account": account_id}
        self._post('jobs/CheckinCredentials', job_data)

    def authenticate(self, username, password, provider, domain):
        auth_data = {
            'username': username,
            'password': password,
            'grant_type': 'password',
            'provider': provider
        }
        req = Request(self._build_url('authorize'), data=urlencode(auth_data).encode())
        response = self._do_request(req)
        if response is None:
            raise PassmanException("Authentication failed")

        return response['access_token'], self._create_auth_token(username, provider, domain)

    def whoami(self):
        return self._get('whoami')

    def _create_auth_token(self, username, provider, domain):
        message = 'un:' + username + ' iss:' + provider + ' dom:' + domain + ' ua:scb rt:' + str(time.time()) + ' '
        message += (16 - len(message) % 16) * '/'
        iv = os.urandom(16)
        cipher = AES.new(AGENT_SECRET, AES.MODE_CBC, iv)
        token_info = cipher.encrypt(message.encode())
        result = AGENT_IDENTIFIER + hashlib.sha256(AGENT_SCB).hexdigest().encode() + binascii.hexlify(token_info) + \
            binascii.hexlify(iv)
        return result.decode()

    def _get(self, resource, parameters=None):
        headers = self._build_authorization_header()
        req = Request(self._build_url(resource, parameters), headers=headers)
        return self._do_request(req)

    def _post(self, resource, post_data, parameters=None):
        headers = self._build_authorization_header()
        headers['content-type'] = 'application/json'
        req = Request(self._build_url(resource, parameters), data=json.dumps(post_data).encode(), headers=headers)
        return self._do_request(req)

    def _do_request(self, req):
        try:
            response = urlopen(req, context=self._get_ssl_context())
            response_body = response.read()
            return json.loads(response_body)
        except HTTPError as exc:
            print("Request failed", exc, exc.fp.read())
            return None

    def _get_ssl_context(self):
        ssl_context = ssl.create_default_context(cadata=self._passman_ca_cert)
        if self._passman_ca_cert is None:
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE
        else:
            ssl_context.verify_mode = ssl.CERT_REQUIRED
            ssl_context.check_hostname = self._passman_check_host_name

        return ssl_context

    def _build_authorization_header(self):
        if self._access_token is None:
            raise PassmanException("Missing access token, needs to authenticate first")

        headers = {
            'authorization': "Bearer {}".format(self._access_token), 'Agent-auth': self._agent_auth
        }
        return headers

    def _build_url(self, resource, parameters=None):
        if parameters is not None:
            params = "/?" + "&".join("{}={}".format(k, v) for k, v in parameters.items())
        else:
            params = ""

        return "{}/api/{}{}".format(self._passman_url, resource, params)


class Plugin:
    def __init__(self, configuration):
        config = ConfigParser(DEFAULT_CONFIG)
        config.read_file(StringIO(configuration))

        self._use_credential = config.get('auth', 'use_credential')
        self._passman_provider = config.get('auth', 'provider')
        self._passman_domain = config.get('auth', 'domain')
        self._passman_username = config.get('auth', 'username')
        self._passman_password = config.get('auth', 'password')
        
        self._passman_username = None if self._passman_username == '' else self._passman_username
        self._passman_password = None if self._passman_password == '' else self._passman_password

        self._passman_group = config.get('search', 'under_group')

        self._passman = Passman(config)

    def remove_password_from_debug(self, kwargs):
        mod_kwargs = {}
        for key, value in kwargs.items():
            if 'password' in key:
                mod_kwargs[key] = '****'
            else:
                mod_kwargs[key] = value
        return mod_kwargs

    def get_password_list(self, session_id, cookie, target_username, target_host, **kwargs):
        if DEBUG:
            print('credstore get_password_list', target_username, target_host, self.remove_password_from_debug(kwargs))

        try:
            credential = self._get_credential(
                cookie, target_username, target_host, 'password', **kwargs)
            return self._send_response(cookie, passwords=[credential, ])
        except PassmanException as exc:
            print("Error checking out password for {}@{}: '{}'".format(target_username, target_host, exc))
            return self._send_response(cookie)

    def get_private_key_list(self, session_id, cookie, target_username, target_host, **kwargs):
        if DEBUG:
            print('credstore get_private_key_list', target_username, target_host,
                  self.remove_password_from_debug(kwargs))

        try:
            credential = self._get_credential(cookie, target_username, target_host, 'ssh-key', **kwargs)

            # ssh key type is hard-coded here...
            return self._send_response(cookie, private_keys=[('ssh-rsa', credential), ])
        except PassmanException as exc:
            print("Error checking out ssh-key for {}@{}: '{}'".format(target_username, target_host, exc))
            return self._send_response(cookie)

    def _get_credential(self, cookie, target_username, target_host, credential_type, **kwargs):
        self._auth_session(cookie, **kwargs)

        account_id = None
        if 'session_cookie' in kwargs:
            account_id = kwargs['session_cookie'].get('account', None)

        if not account_id:
            try:
                account = self._passman.get_account(target_username, target_host, self._passman_group)
            except PassmanException:
                if self._is_ip_address(target_host):
                    if DEBUG:
                        print('target host is ip address, try with fqdn')
                    target_fqdn = self._get_fqdn(target_host)
                    if DEBUG:
                        print('target fqdn:', target_fqdn)
                    account = self._passman.get_account(target_username, target_fqdn, self._passman_group)
                    if DEBUG:
                        print('Found account', account)
                else:
                    raise
        credential, is_exclusive = self._passman.checkout_credential(account, credential_type)
        if DEBUG:
            print("Found {} for {}@{}".format(credential_type, target_username, target_host))

        cookie['account'] = account['id']
        cookie['is_exclusive'] = is_exclusive
        return credential

    @staticmethod
    def _is_ip_address(target_host):
        # simplification
        return all([ch.isdigit() or ch == '.' for ch in target_host])

    @staticmethod
    def _get_fqdn(ip_address):
        addr = reversename.from_address(ip_address)
        fqdn = str(resolver.query(addr, "PTR")[0])
        if fqdn.endswith('.'):
            fqdn = fqdn[:-1]
        return fqdn.strip()

    def authentication_completed(self, session_id, cookie):
        return self._send_response(cookie)

    def session_ended(self, session_id, cookie):
        try:
            self._checkin_credentials(cookie)
        except PassmanException as exc:
            print("Error checking in credential {}".format(exc))
        return self._send_response(cookie)

    def _auth_session(self, cookie, **kwargs):
        access_token, auth_token = self._get_access_token(**kwargs)
        cookie['access_token'] = access_token
        cookie['auth_token'] = auth_token
        self._passman.set_access_token(access_token)
        self._passman.set_auth_token(auth_token)

    def _get_access_token(self, **kwargs):
        if 'session_cookie' in kwargs and 'token' in kwargs['session_cookie']:
            print('Access token is used for authentication')
            return kwargs['session_cookie']['token']

        print("no access token given, let's authenticate")
        username, password = self._get_passman_credential(**kwargs)
        return self._passman.authenticate(username, password, self._passman_provider, self._passman_domain)

    def _get_passman_credential(self, **kwargs):
        if self._use_credential == 'explicit':
            if self._passman_username is not None and self._passman_password is not None:
                print("Using explictly configured proxy user to authenticate")
                return self._passman_username, self._passman_password
            else:
                raise PassmanException('Missing username and password')
        elif self._use_credential == 'gateway':
            if 'gateway_username' in kwargs and 'gateway_password' in kwargs and \
                    kwargs['gateway_username'] and kwargs['gateway_password']:
                print("Using gateway user to authenticate")
                return kwargs['gateway_username'], kwargs['gateway_password']
            else:
                raise PassmanException('Missing gateway credential')
        else:
            raise PassmanException('Invalid credential source: {}'.format(self._use_credential))

    def _checkin_credentials(self, cookie):
        if 'is_exclusive' not in cookie or not cookie['is_exclusive']:
            return

        if 'account' not in cookie:
            raise PassmanException('Missing account')

        if 'access_token' not in cookie:
            raise PassmanException('Missing access token')

        self._passman.set_access_token(cookie['access_token'])
        print("Checkin credentials")
        self._passman.checkin_credentials(cookie['account'])

    @staticmethod
    def _send_response(cookie, passwords=None, private_keys=None):
        result = {
            'cookie': cookie
        }

        if passwords is not None:
            result['passwords'] = passwords

        if private_keys is not None:
            result['private_keys'] = private_keys
        return result


if __name__ == '__main__':
    import sys
    plugin = Plugin(open(sys.argv[1], 'r').read())

    session_id = 'session_id'
    if len(sys.argv) == 6:
        gw_user = sys.argv[4]
        gw_pass = sys.argv[5]
    else:
        gw_user = None
        gw_pass = None

    # method = plugin.get_private_key_list
    method = plugin.get_password_list

    result = method(target_host=sys.argv[2], target_username=sys.argv[3], session_id=session_id,
                    cookie={}, gateway_username=gw_user, gateway_password=gw_pass, session_cookie={})
    print('RESULT', result)
    result = plugin.authentication_completed(session_id=session_id, cookie=result['cookie'])
    result = plugin.session_ended(session_id=session_id, cookie=result['cookie'])
